<?php

namespace Drupal\search_api_nuclia\Plugin\search_api\backend;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\search_api\Backend\BackendPluginBase;
use Drupal\search_api\IndexInterface;
use Drupal\search_api\Plugin\PluginFormTrait;
use Drupal\search_api\Query\QueryInterface;
use Drupal\search_api_nuclia\Service\DebugLogStacker;
use Drupal\search_api_nuclia\Service\NucliadbVsDrupalMap;
use Drupal\search_api_nuclia\NucliaRequests;
use Drupal\search_api_nuclia\Service\NucliadbHelper;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Contracts\HttpClient\Exception\HttpExceptionInterface;

/**
 * Search api backend for NucliaDB.
 *
 * @SearchApiBackend(
 *   id = "search_api_nuclia",
 *   label = @Translation("NucliaDB"),
 *   description = @Translation("Index items using a NucliaDB Search.")
 * )
 */
class SearchApiNucliadbBackend extends BackendPluginBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected LanguageManagerInterface $languageManager;

  /**
   * The Config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected ConfigFactoryInterface $configFactory;

  /**
   * The Nuclia helper.
   *
   * @var \Drupal\search_api_nuclia\Service\NucliadbHelper
   */
  protected NucliadbHelper $nucliadbHelper;

  /**
   * The Nuclia helper.
   *
   * @var \Drupal\search_api_nuclia\Service\NucliadbVsDrupalMap
   */
  protected NucliadbVsDrupalMap $nucliadbVsDrupalMap;

  /**
   * Debug log stacker.
   *
   * @var \Drupal\search_api_nuclia\Service\DebugLogStacker
   */
  protected DebugLogStacker $debugLogStacker;

  /**
   * Nuclia requests.
   *
   * @var \Drupal\search_api_nuclia\NucliaRequests
   */
  protected NucliaRequests $nucliaRequests;

  /**
   * Constructor.
   */
  public function __construct(
        array $configuration,
        $pluginId,
        array $pluginDefinition,
        LanguageManagerInterface $languageManager,
        ConfigFactoryInterface $configFactory,
        NucliadbHelper $nucliadbHelper,
        NucliadbVsDrupalMap $nucliadbVsDrupalMap,
        DebugLogStacker $debugLogStacker
    ) {
    parent::__construct($configuration, $pluginId, $pluginDefinition);
    $this->languageManager = $languageManager;
    $this->configFactory = $configFactory;
    $this->nucliadbHelper = $nucliadbHelper;
    $this->nucliadbVsDrupalMap = $nucliadbVsDrupalMap;
    $this->debugLogStacker = $debugLogStacker;
    $this->debugLogStacker->setIsEnabled($this->getLogDebug());
    $this->setLogger($this->debugLogStacker->getLogger());
    if (!empty($this->getZone()) && !empty($this->getToken()) && !empty($this->getKbid())) {
      $this->nucliaRequests = new NucliaRequests(
            $this->getZone(), $this->getToken(),
            $this->getKbid(), $this->logger, $this->debugLogStacker, $this->nucliadbHelper
        );
    }

  }

  /**
   * {@inheritDoc}
   */
  public function supportsDataType($type) {
    if ($type === "nuclia_file") {
      return TRUE;
    }

    // This would normally disable boolean data type for Search API NucliaDB
    // Backend (which doesn't really make sense), but it unfortunately doesn't
    // seem to work. A workaround had been implemented to ignore this type.
    // @see NucliadbHelper::prepareNucliaResourceInfo
    if ($type === "boolean") {
      return FALSE;
    }

    // @todo Change the autogenerated stub.
    return parent::supportsDataType($type);
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $pluginId, $pluginDefinition): SearchApiNucliadbBackend {
    /** @var \Drupal\Core\Language\LanguageManagerInterface $languageManager */
    $languageManager = $container->get('language_manager');
    /** @var \Drupal\Core\Config\ConfigFactoryInterface $configFactory */
    $configFactory = $container->get('config.factory');
    /** @var \Drupal\search_api_nuclia\Service\NucliadbHelper $nucliaDbHelper */
    $nucliaDbHelper = $container->get('search_api_nuclia.helper');
    /** @var \Drupal\search_api_nuclia\Service\NucliadbVsDrupalMap $nucliadbVsDrupalMap */
    $nucliadbVsDrupalMap = $container->get('search_api_nuclia.nucliadb_vs_drupal_map');
    /** @var \Drupal\search_api_nuclia\Service\DebugLogStacker $debugLogStacker */
    $debugLogStacker = $container->get('search_api_nuclia.debug_log_stacker');

    $backend = new static(
          $configuration,
          $pluginId,
          $pluginDefinition,
          $languageManager,
          $configFactory,
          $nucliaDbHelper,
          $nucliadbVsDrupalMap,
          $debugLogStacker
      );

    return $backend;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'zone' => '',
      'token' => '',
      'kbid' => '',
      'process_public_files_as_link' => TRUE,
      'log_debug' => FALSE,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['help'] = [
      '#markup' => '<p>' . $this->t(
          'The zone, token, kbid, knowledge base id can be found or configured at your Nuclia cloud account. Please sign up at <a href="@link_signup" target="blank">@link_signup</a> and sign in at <a href="@link_login" target="blank">@link_login</a>',
          [
            '@link_login' => 'https://nuclia.cloud/user/login',
            '@link_signup' => 'https://nuclia.com/sign-up/',
          ]
      ) . '</p>',
    ];

    $form['zone'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Zone'),
      '#description' => $this->t('The zone from your NucliaDB account.'),
      '#default_value' => $this->getZone(),
      '#required' => TRUE,
      '#size' => 60,
      '#maxlength' => 128,
    ];

    $form['token'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Token'),
      '#description' => $this->t('The Token you have created in your NucliaDB account.'),
      '#default_value' => $this->getToken(),
      '#required' => TRUE,
    ];

    $form['kbid'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Knowledge base Identifier'),
      '#description' => $this->t('Your NucliaDB Knowledge base Identifier.'),
      '#default_value' => $this->getKbid(),
      '#required' => TRUE,
      '#size' => 60,
      '#maxlength' => 128,
    ];

    $form['process_public_files_as_link'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Process public files as link'),
      '#description' => $this->t('Allow to process public files as link which is equivalent for NucliaDB.'),
      '#default_value' => $this->getProcessPublicFilesAsLink(),
    ];

    $form['log_debug'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Log debugging traces.'),
      '#description' => $this->t('Log debugging traces into the enabled drupal logger if any.'),
      '#default_value' => $this->getLogDebug(),
    ];

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function indexItems(IndexInterface $index, array $items) {
    $this->debugLogStacker->start();
    $indexedIds = [];

    // Get existing resource rids.
    $existingNucliadbRids = $this->nucliadbVsDrupalMap->getExistingRidsFromSearchItems($items);
    $this->debugLogStacker->appendSection(
          'Items to index info',
          [
          [
            'name' => 'Item id to index',
            'value' => json_encode(
              array_keys($items),
              JSON_PRETTY_PRINT
            ),
            'type' => 'json',
          ],
          [
            'name' => 'Found indexes matching rids',
            'value' => json_encode(
              $existingNucliadbRids,
              JSON_PRETTY_PRINT
            ),
            'type' => 'json',
          ],
          ],
          DebugLogStacker::formatContext(__FUNCTION__, $this)
      );
    /** @var \Drupal\search_api\Item\ItemInterface $item */
    foreach ($items as $itemId => $item) {
      $rid = NULL;
      $checksums = [];

      // Determine if the current item match an existing resource.
      if (array_key_exists($itemId, $existingNucliadbRids)) {
        $rid = $existingNucliadbRids[$itemId]->{NucliadbVsDrupalMap::RID_COL_NAME};
        // Get current resource's fields md5 checksums.
        $checksums = $this->nucliaRequests->getResourceFieldsChecksums($rid);
      }

      // Prepare resource info to submit resource to Nuclia.
      ['body' => $resourceBody, 'files' => $resourceFiles]
        = $this->nucliadbHelper->prepareNucliaResourceInfo($item, $this->getProcessPublicFilesAsLink(), $checksums);

      try {

        // Activate NucliaDB synchronous mode if some files have to be attached
        // to the resource just after creating/updating it : Without
        // synchronous mode activated, NucliaDB send a success response on
        // resource create/update before the resource is effectively available.
        // Creating files as field of that resource just right after then fail.
        if (count($resourceFiles)) {
          $this->nucliaRequests->setSynchronizedMode(TRUE);
        }

        // NucliaDB Resource update case.
        if ($rid !== NULL) {
          // Retrieve the current resource rid.
          $rid = $existingNucliadbRids[$itemId]->{NucliadbVsDrupalMap::RID_COL_NAME};

          // Update the NucliaDB side resource.
          $this->nucliaRequests->updateResource($rid, $resourceBody);
        }

        // Create NucliaDB Resource.
        else {

          // Update the NucliaDB side resource.
          $rid = $this->nucliaRequests->createResource($resourceBody);

          if ($rid !== NULL) {
            // Feed the Nucliadb vs Drupal map table.
            $this->nucliadbVsDrupalMap->createNucliadbDrupalMapInfo($itemId, $rid);
          }
        }

        // Stop NucliaDB synchronous mode.
        if (count($resourceFiles)) {
          $this->nucliaRequests->setSynchronizedMode(FALSE);
        }

        // Upload files if needed.
        if ($rid !== NULL) {
          foreach ($resourceFiles as $file) {
            ['name' => $name, 'data' => $data, 'md5' => $md5] = $file;
            $this->nucliaRequests->addFileField($rid, $name, $data, $md5);
          }
        }
      }
      catch (HttpExceptionInterface $exception) {
        $this->logger->error(
          'Http exception occurred when calling NucilaDB api : <br> @exception', [
            '@exception' => (string) $exception,
          ]
        );

        // If a http exception occurs, items loop breaks here avoiding setting
        // the current item id as indexed.
        continue;
      }

      // The item ID indexation is successfully done.
      $indexedIds[] = $itemId;
    }
    $this->debugLogStacker->appendSection(
          'Return indexed id',
          [
          [
            'name' => 'Indexed ids',
            'value' => json_encode($indexedIds, JSON_PRETTY_PRINT),
            'type' => 'json',
          ],
          ],
          DebugLogStacker::formatContext(__FUNCTION__, $this)
      );
    $this->debugLogStacker->stop();
    return $indexedIds;
  }

  /**
   * {@inheritDoc}
   */
  public function deleteItems(IndexInterface $index, array $itemIds) {
    $this->debugLogStacker->start();
    // Make a list of Nuclia rids matching the search items id.
    $existingRids = $this->nucliadbVsDrupalMap->getExistingRidsFromSearchItemIds($itemIds, [NucliadbVsDrupalMap::RID_COL_NAME], 0, 'col');

    // Remove resources from Nuclia.
    if (count($existingRids) > 0) {
      foreach ($existingRids as $rid) {
        try {
          $this->nucliaRequests->deleteResource($rid);
        }
        catch (\Exception $e) {
          $this->logger->warning($e->getMessage());
        }
      }

      // Remove Drupal/NucliaDB mapping info.
      $this->nucliadbVsDrupalMap->deleteNucliadbDrupalMapInfoFromNucliadbRids($existingRids);
    }
    $this->debugLogStacker->stop();
  }

  /**
   * {@inheritDoc}
   */
  public function deleteAllIndexItems(IndexInterface $index, $datasourceId = NULL) {
    $this->debugLogStacker->start();
    $searchItemInfo = $this->nucliadbVsDrupalMap->getAllSearchItemInfo();
    foreach ($searchItemInfo as $searchItemInfoItem) {
      try {
        $this->nucliaRequests->deleteResource($searchItemInfoItem->nucliadb_rid);
      }
      catch (\Exception $e) {
        $this->logger->warning($e->getMessage());
      }
    }
    $this->nucliadbVsDrupalMap->deleteAll();
    $this->debugLogStacker->stop();
  }

  /**
   * {@inheritDoc}
   */
  public function search(QueryInterface $query) {
    $this->debugLogStacker->start();

    $results = $query->getResults();
    $options = $query->getOptions();
    $search_api_index = $query->getIndex();

    ['limit' => $limit, 'offset' => $offset] = $options;
    if ($limit > 0) {
      $pageNumber = $offset / $limit;
    }
    else {
      $pageNumber = 1;
    }
    $this->debugLogStacker->appendSection(
          'search',
          [
          ['name' => 'keys', 'value' => $query->getOriginalKeys() ?? ''],
          ['name' => 'limit', 'value' => $limit],
          ['name' => 'offset', 'value' => $offset],
          ],
          DebugLogStacker::formatContext(__FUNCTION__, $this)
      );
    $searchText = $query->getOriginalKeys();
    if (!empty($searchText)) {
      ['result_rids' => $resultRids, 'next_page' => $haveNextPage] = $this->nucliaRequests->search($searchText, $limit, $pageNumber);

      $drupalKeys = $this->nucliadbVsDrupalMap->getSearchItemIdsFromRids($resultRids);

      // Calculate the total results count : This is not the real total result
      // count : NucliaDB can't for now calculate the total amount of results.
      // The only reliable information is that there is a next page.
      // So we calculate a fake total result count to make the views pager
      // displays next pages links.
      $calculatedResultCount = $offset + $limit + ($haveNextPage ? 1 : 0);
      $results->setResultCount($calculatedResultCount);

      // Add item to search result.
      foreach ($drupalKeys as $index => $key) {
        $item = $this->getFieldsHelper()->createItem($search_api_index, $key);
        $results->addResultItem($item);
        if ($index + 1 >= $limit) {
          break;
        }
      }

      // Debug logging.
      $this->debugLogStacker->appendSection(
            'search',
            [
            [
              'name' => 'rids',
              'value' => json_encode($resultRids, JSON_PRETTY_PRINT),
              'type' => 'json',
            ],
            [
              'name' => 'drupal_keys',
              'value' => json_encode($drupalKeys, JSON_PRETTY_PRINT),
              'type' => 'json',
            ],
            ],
            DebugLogStacker::formatContext(__FUNCTION__, $this)
        );
    }
    $this->debugLogStacker->stop();

    return $results;
  }

  /**
   * Retrieves the logger to use.
   *
   * @return \Psr\Log\LoggerInterface
   *   The logger to use.
   */
  public function getLogger() {
    return $this->logger;
  }

  /**
   * Sets the logger to use.
   *
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger to use.
   *
   * @return $this
   *   current object for chaining
   */
  public function setLogger(LoggerInterface $logger) {
    $this->logger = $logger;
    return $this;
  }

  /**
   * Get NucliaDB zone.
   *
   * @return string
   *   Configuration value
   */
  protected function getZone(): string {
    return $this->configuration['zone'];
  }

  /**
   * Get NucliaDB token.
   *
   * @return string
   *   Configuration value
   */
  protected function getToken(): string {
    return $this->configuration['token'];
  }

  /**
   * Get NucliaDB knowledge base id.
   *
   * @return string
   *   Configuration value
   */
  protected function getKbid(): string {
    return $this->configuration['kbid'];
  }

  /**
   * Get public-files-as-link configuration value.
   *
   * @return string
   *   Configuration value
   */
  protected function getProcessPublicFilesAsLink(): string {
    return $this->configuration['process_public_files_as_link'];
  }

  /**
   * Get log debug configuration value.
   *
   * @return string
   *   Configuration value
   */
  protected function getLogDebug(): string {
    return $this->configuration['log_debug'] ?? FALSE;
  }

}
